import { assert } from 'chai'
import axios from 'axios'
import { pick } from 'lodash'
import sinon from 'sinon'
import { createPinia, setActivePinia } from 'pinia'
import { mutations } from '@/store/process.js'
import {
  artifactSample,
  workerConfigurationsSample,
  workerVersionsSample,
  workerRunSample,
  workerRunsSample,
  processSample,
  processesSample,
  processElementsSample,
  taskSample
} from '../samples.js'
import store from './index.spec.js'
import { assertRejects, FakeAxios } from '../testhelpers.js'
import { useWorkerStore } from '@/stores'

describe('process', () => {
  let sandbox

  before('Setup Sinon sandbox', () => {
    sandbox = sinon.createSandbox()
  })

  beforeEach('Setup timeout mock', () => {
    sandbox.stub(global, 'setTimeout')
    sandbox.stub(global, 'clearTimeout')
  })

  afterEach('Tear down timeout mock', () => {
    sandbox.restore()
  })

  describe('mutations', () => {
    it('setProcessWorkerRuns', () => {
      const state = { processWorkerRuns: {} }
      mutations.setProcessWorkerRuns(state, { processId: 'processid', workerRuns: workerRunsSample })
      assert.deepStrictEqual(state, {
        processWorkerRuns: {
          processid: {
            workerRun1: workerRunsSample[0],
            workerRun2: workerRunsSample[1],
            workerRun3: workerRunsSample[2]
          }
        }
      })
    })

    it('removeProcessWorkerRun', () => {
      const state = {
        processWorkerRuns: {
          processid: {
            workerRun1: {
              id: 'workerRun1',
              parents: [],
              version: 'versionid1'
            },
            workerRun2: {
              id: 'workerRun2',
              parents: ['workerRun1'],
              version: 'versionid2'
            }
          }
        }
      }
      mutations.removeProcessWorkerRun(state, { processId: 'processid', workerRunId: 'workerRun1' })
      assert.deepStrictEqual(state, {
        processWorkerRuns: {
          processid: {
            workerRun2: {
              id: 'workerRun2',
              parents: [],
              version: 'versionid2'
            }
          }
        }
      })
    })

    it('setProcesses', () => {
      const state = {
        processes: {
          process_0: { id: 'process_0', _complete: false },
          process_1: { id: 'process_1', _complete: true }
        }
      }
      mutations.setProcesses(state, [
        { id: 'process_0' },
        { id: 'process_1', name: 'P1' },
        { id: 'process_2' }
      ])
      assert.deepStrictEqual(state, {
        processes: {
          process_0: { id: 'process_0', _complete: false },
          process_1: { id: 'process_1', name: 'P1', _complete: true },
          process_2: { id: 'process_2', _complete: false }
        }
      })
    })

    describe('setProcessSets', () => {
      it('sets dataset sets on a process', () => {
        const state = {
          processSets: {
            process2: [
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'test'
              }
            ]
          }
        }
        const processSets = [
          {
            id: 'processsetid2',
            dataset: { id: 'datasetid' },
            set_name: 'validation'
          },
          {
            id: 'processsetid3',
            dataset: { id: 'datasetid2' },
            set_name: 'test'
          }
        ]
        mutations.setProcessSets(state, {
          processId: 'process1',
          results: processSets
        })
        assert.deepStrictEqual(state, {
          processSets: {
            process1: processSets,
            process2: [
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'test'
              }
            ]
          }
        })
      })

      it('adds sets to an existing list of process sets', () => {
        const state = {
          processSets: {
            process1: [
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'test'
              }
            ]
          }
        }
        mutations.setProcessSets(state, {
          processId: 'process1',
          results: [
            {
              id: 'processsetid2',
              dataset: { id: 'datasetid' },
              set_name: 'validation'
            }
          ]
        })
        assert.deepStrictEqual(state, {
          processSets: {
            process1: [
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'test'
              },
              {
                id: 'processsetid2',
                dataset: { id: 'datasetid' },
                set_name: 'validation'
              }
            ]
          }
        })
      })

      it('deduplicates', () => {
        const state = {
          processSets: {
            process1: [
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'test'
              }
            ]
          }
        }
        mutations.setProcessSets(state, {
          processId: 'process1',
          results: [
            {
              id: 'processsetid',
              dataset: { id: 'datasetid' },
              set_name: 'test'
            },
            {
              id: 'processsetid2',
              dataset: { id: 'datasetid' },
              set_name: 'validation'
            },
            {
              id: 'processsetid',
              dataset: { id: 'datasetid' },
              set_name: 'test'
            },
            {
              id: 'processsetid2',
              dataset: { id: 'datasetid' },
              set_name: 'validation'
            }
          ]
        })
        assert.deepStrictEqual(state, {
          processSets: {
            process1: [
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'test'
              },
              {
                id: 'processsetid2',
                dataset: { id: 'datasetid' },
                set_name: 'validation'
              }
            ]
          }
        })
      })
    })

    describe('removeProcessSet', () => {
      it('removes a set from a process', () => {
        const state = {
          processSets: {
            process1: [
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'test'
              },
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'validation'
              }
            ]
          }
        }
        mutations.removeProcessSet(state, {
          processId: 'process1',
          processSetId: 'processsetid'
        })
        assert.deepStrictEqual(state, {
          processSets: {
            process1: [
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'validation'
              }
            ]
          }
        })
      })

      it('ignores a set that is already removed', () => {
        const state = {
          processSets: {
            process1: [
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'validation'
              }
            ]
          }
        }
        mutations.removeProcessSet(state, {
          processId: 'process1',
          processSetId: 'processsetid2'
        })
        assert.deepStrictEqual(state, {
          processSets: {
            process1: [
              {
                id: 'processsetid',
                dataset: { id: 'datasetid' },
                set_name: 'validation'
              }
            ]
          }
        })
      })
    })

    it('setProcessPage', () => {
      const state = {
        processPage: { count: 1, results: ['process_0'] }
      }
      mutations.setProcessPage(state, { count: 2, results: [{ id: 'process_1' }, { id: 'process_2' }] })
      assert.deepStrictEqual(state, {
        processPage: {
          count: 2,
          results: ['process_1', 'process_2']
        }
      })
    })

    it('removeProcess', () => {
      const state = {
        processes: {
          process_0: { id: 'process_0' },
          process_1: { id: 'process_1' }
        }
      }
      mutations.removeProcess(state, 'process_1')
      assert.deepStrictEqual(state, {
        processes: { process_0: { id: 'process_0' } }
      })
    })

    it('setProcessElementsPage', () => {
      const state = {
        processElementsPage: {
          process_0: { count: 42, page: 1, results: [] },
          process_1: { count: 1337, results: [{ id: 'elt_id', type: 'page' }] }
        }
      }
      mutations.setProcessElementsPage(state, { processId: 'process_1', response: processElementsSample })
      assert.deepStrictEqual(state, {
        processElementsPage: {
          process_0: { count: 42, page: 1, results: [] },
          // Count is not erased by the processElements payload
          process_1: { count: 1337, ...processElementsSample }
        }
      })
    })

    describe('setProcess', () => {
      it('sets the process and tasks', () => {
        const state = { processes: {}, tasks: {} }
        const expectedProcess = {
          ...processSample,
          _complete: true
        }
        delete expectedProcess.tasks

        mutations.setProcess(state, processSample)

        assert.deepStrictEqual(state, {
          processes: {
            [processSample.id]: expectedProcess
          },
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: null,
              process_id: processSample.id
            }
          }
        })
      })

      it('updates existing tasks', () => {
        assert.notStrictEqual(taskSample.state, 'unscheduled')
        const state = {
          processes: {},
          tasks: {
            [taskSample.id]: {
              id: taskSample.id,
              state: 'unscheduled'
            }
          }
        }

        const expectedProcess = {
          ...processSample,
          _complete: true
        }
        delete expectedProcess.tasks

        mutations.setProcess(state, processSample)
        assert.deepStrictEqual(state, {
          processes: {
            [processSample.id]: expectedProcess
          },
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: null,
              process_id: processSample.id
            }
          }
        })
        assert.ok(clearTimeout.notCalled)
      })

      it('stops the task polling before removing tasks', () => {
        const state = {
          processes: {},
          tasks: {
            deadtask: {
              id: 'deadtask',
              state: 'error',
              timeoutId: null
            },
            polledtask: {
              id: 'polledtask',
              state: 'running',
              timeoutId: 42
            }
          }
        }
        const expectedProcess = {
          ...processSample,
          _complete: true
        }
        delete expectedProcess.tasks

        mutations.setProcess(state, processSample)

        assert.deepStrictEqual(state, {
          processes: {
            [processSample.id]: expectedProcess
          },
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: null,
              process_id: processSample.id
            }
          }
        })
        assert.ok(clearTimeout.calledOnceWithExactly(42))
      })

      it('detects incomplete processes', () => {
        const state = { processes: {}, tasks: {} }

        const payload = {
          ...processSample
        }
        delete payload.farm

        const expectedProcess = {
          ...processSample,
          _complete: false
        }
        delete expectedProcess.farm
        delete expectedProcess.tasks

        mutations.setProcess(state, payload)

        assert.deepStrictEqual(state, {
          processes: {
            [processSample.id]: expectedProcess
          },
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: null,
              process_id: processSample.id
            }
          }
        })
      })

      it('preserves existing attributes of incomplete processes', () => {
        const state = {
          processes: {
            processid: {
              ...processSample,
              _complete: true
            }
          },
          tasks: {
            [taskSample.id]: taskSample
          }
        }
        delete state.processes.processid.tasks

        // We will call setProcesses with many missing attributes and an updated state
        const payload = {
          id: 'processid',
          state: 'stopped'
        }
        delete payload.tasks

        const expectedProcess = {
          // Everything on the previous process is kept
          ...processSample,
          // State will be updated
          state: 'stopped',
          // Process stays complete despite the lack of tasks
          _complete: true
        }
        delete expectedProcess.tasks

        mutations.setProcess(state, payload)

        assert.deepStrictEqual(state, {
          processes: {
            [processSample.id]: expectedProcess
          },
          // Tasks are unchanged too
          tasks: {
            [taskSample.id]: taskSample
          }
        })
      })
    })

    it('setPollingProcessId', () => {
      const state = { pollingProcessId: null }
      mutations.setPollingProcessId(state, 'processid')
      assert.deepStrictEqual(state, { pollingProcessId: 'processid' })
    })

    it('setProcessTimeoutId', () => {
      const state = { processTimeoutId: null }
      mutations.setProcessTimeoutId(state, 1234)
      assert.deepStrictEqual(state, { processTimeoutId: 1234 })
    })

    it('setTask', () => {
      const state = {
        workflow: { id: 'workflowid' },
        tasks: {}
      }
      const updatedTask = { ...taskSample, state: 'stopping' }
      mutations.setTask(state, updatedTask)
      assert.deepStrictEqual(state, {
        workflow: { id: 'workflowid' },
        tasks: {
          [updatedTask.id]: {
            ...updatedTask,
            timeoutId: null
          }
        }
      })
    })

    it('setTaskTimeoutId', () => {
      const state = { tasks: { [taskSample.id]: { ...taskSample } } }
      mutations.setTaskTimeoutId(state, { taskId: taskSample.id, timeoutId: 1234 })
      assert.deepStrictEqual(state, {
        tasks: {
          [taskSample.id]: {
            ...taskSample,
            timeoutId: 1234
          }
        }
      })
    })

    it('setTaskArtifacts', () => {
      const state = { artifacts: {} }
      mutations.setTaskArtifacts(state, { taskId: taskSample.id, artifacts: [artifactSample] })
      assert.deepStrictEqual(state, {
        artifacts: {
          [taskSample.id]: [artifactSample]
        }
      })
    })

    describe('stopPolling', () => {
      it('stops workflow and task polling', () => {
        const state = {
          pollingProcessId: 'processid',
          processTimeoutId: 1234,
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: 42
            }
          }
        }
        mutations.stopPolling(state)
        assert.deepStrictEqual(state, {
          pollingProcessId: null,
          processTimeoutId: null,
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: null
            }
          }
        })
        assert.ok(clearTimeout.calledTwice)
        assert.ok(clearTimeout.firstCall.calledWithExactly(1234))
        assert.ok(clearTimeout.secondCall.calledWithExactly(42))
      })

      it('handles an already stopped polling', () => {
        const state = {
          pollingProcessId: 'processid',
          processTimeoutId: null,
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: null
            }
          }
        }
        mutations.stopPolling(state)
        assert.deepStrictEqual(state, {
          pollingProcessId: null,
          processTimeoutId: null,
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: null
            }
          }
        })
        assert.ok(clearTimeout.notCalled)
      })
    })

    describe('stopTaskPolling', () => {
      it('stops task polling', () => {
        const state = {
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: 1234
            }
          }
        }
        mutations.stopTaskPolling(state, taskSample.id)
        assert.ok(clearTimeout.calledOnceWithExactly(1234))
        assert.deepStrictEqual(state, {
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: null
            }
          }
        })
      })

      it('handles an already stopped polling', () => {
        const state = {
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: null
            }
          }
        }
        mutations.stopTaskPolling(state, taskSample.id)
        assert.ok(clearTimeout.notCalled)
        assert.deepStrictEqual(state, {
          tasks: {
            [taskSample.id]: {
              ...taskSample,
              timeoutId: null
            }
          }
        })
      })

      // Some tasks can go missing before the polling stops, for example if the store gets reset due to logging out.
      it('ignores non-existent tasks', () => {
        const state = { tasks: {} }
        mutations.stopTaskPolling(state, 'aaaaaaaaaa')
        assert.ok(clearTimeout.notCalled)
        assert.deepStrictEqual(state, { tasks: {} })
      })
    })

    it('reset', () => {
      const state = {
        processWorkerRuns: workerRunsSample,
        processes: processesSample,
        processElementsPage: processElementsSample,
        processSets: {
          processid: [
            {
              id: 'processsetid',
              dataset: { id: 'datasetid' },
              set_name: 'test'
            }
          ]
        },
        tasks: { [taskSample.id]: taskSample },
        pollingProcessId: 'processid',
        processTimeoutId: 1234
      }
      mutations.reset(state)
      assert.deepStrictEqual(state, {
        processWorkerRuns: {},
        processSets: {},
        processes: {},
        processPage: {},
        processElementsPage: {},
        pollingProcessId: null,
        processTimeoutId: null,
        tasks: {},
        artifacts: {}
      })
    })
  })

  describe('actions', () => {
    let mock, workerStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      workerStore = useWorkerStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.reset()
      workerStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('listWorkerRuns', () => {
      it('lists worker runs for a specific process', async () => {
        const replyResults = [workerRunsSample[0], workerRunsSample[1]]

        mock.onGet('/process/processid/workers/').reply(200, { number: 1, next: null, results: replyResults })
        await store.dispatch('process/listWorkerRuns', { processId: 'processid', page: 1 })
        assert.deepStrictEqual(store.history, [
          { action: 'process/listWorkerRuns', payload: { processId: 'processid', page: 1 } },
          { mutation: 'process/setProcessWorkerRuns', payload: { processId: 'processid', workerRuns: replyResults } }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/process/processid/workers/'
          }
        ])
        assert.deepStrictEqual(store.state.process.processWorkerRuns, { processid: { workerRun1: replyResults[0], workerRun2: replyResults[1] } })
        assert.deepStrictEqual(
          workerStore.workerConfigurations,
          {
            workerid1: {
              configid1: workerConfigurationsSample[0]
            },
            workerid2: {
              configid2: workerConfigurationsSample[1]
            }
          }
        )
      })
    })

    describe('createWorkerRun', () => {
      it('creates a worker run on a process', async () => {
        const reply = {
          id: 'workerRun2',
          parents: ['workerRun1'],
          version: workerVersionsSample.results[1].id
        }
        mock.onPost('/process/processid/workers/').reply(200, reply)
        await store.dispatch('process/createWorkerRun', {
          processId: 'processid',
          workerRun: {
            parents: ['workerRun1'],
            version: workerVersionsSample.results[1].id
          }
        })
        assert.deepStrictEqual(store.history, [
          {
            action: 'process/createWorkerRun',
            payload: {
              processId: 'processid',
              workerRun: {
                parents: ['workerRun1'],
                version: workerVersionsSample.results[1].id
              }
            }
          },
          {
            mutation: 'process/setProcessWorkerRuns',
            payload: {
              processId: 'processid',
              workerRuns: [reply]
            }
          }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'post',
            url: '/process/processid/workers/',
            data: {
              parents: ['workerRun1'],
              version: workerVersionsSample.results[1].id
            }
          }
        ])
        assert.deepStrictEqual(store.state.process.processWorkerRuns, { processid: { workerRun2: reply } })
      })
    })

    describe('updateWorkerRun', () => {
      it('updates a worker run from a process', async () => {
        store.state.process.processWorkerRuns = {
          processid: {
            workerRun2: {
              id: 'workerRun2',
              parents: ['workerRun1'],
              version: workerVersionsSample.results[1].id
            }
          }
        }

        const reply = {
          id: 'workerRun2',
          parents: [],
          version: workerVersionsSample.results[1].id
        }
        mock.onPatch('/process/workers/workerRun2/').reply(200, reply)

        const payload = { parents: [] }
        await store.dispatch('process/updateWorkerRun', { processId: 'processid', workerRunId: 'workerRun2', payload })
        assert.deepStrictEqual(store.history, [
          { action: 'process/updateWorkerRun', payload: { processId: 'processid', workerRunId: 'workerRun2', payload } },
          {
            mutation: 'process/setProcessWorkerRuns',
            payload: {
              processId: 'processid',
              workerRuns: [reply]
            }
          }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'patch',
            url: '/process/workers/workerRun2/',
            data: {
              parents: []
            }
          }
        ])
        assert.deepStrictEqual(store.state.process.processWorkerRuns, { processid: { workerRun2: reply } })
      })

      it('handles errors', async () => {
        const payload = { parents: [] }
        mock.onPatch('/process/workers/workerRun2/').reply(500, { detail: ['Something went wrong'] })
        const error = await assertRejects(async () => store.dispatch('process/updateWorkerRun', { processId: 'processid', workerRunId: 'workerRun2', payload }))
        assert.strictEqual(error.toString(), 'Error: Request failed with status code 500')
        assert.deepStrictEqual(store.history, [
          { action: 'process/updateWorkerRun', payload: { processId: 'processid', workerRunId: 'workerRun2', payload } }
        ])
      })
    })

    describe('deleteWorkerRun', () => {
      it('deletes a worker run from a process', async () => {
        store.state.process.processWorkerRuns = {
          processid: {
            workerRun2: {
              id: 'workerRun2',
              parents: ['workerRun1'],
              version: workerVersionsSample.results[1].id
            }
          }
        }

        mock.onDelete('/process/workers/workerRun2/').reply(204)
        await store.dispatch('process/deleteWorkerRun', { processId: 'processid', workerRunId: 'workerRun2' })
        assert.deepStrictEqual(store.history, [
          { action: 'process/deleteWorkerRun', payload: { processId: 'processid', workerRunId: 'workerRun2' } },
          {
            mutation: 'process/removeProcessWorkerRun',
            payload: {
              processId: 'processid',
              workerRunId: 'workerRun2'
            }
          }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'delete',
            url: '/process/workers/workerRun2/'
          }
        ])
        assert.deepStrictEqual(store.state.process.processWorkerRuns, { processid: {} })
      })

      it('deletes a worker run from a process and its references on its childen', async () => {
        store.state.process.processWorkerRuns = {
          processid: {
            workerRun1: {
              id: 'workerRun1',
              parents: [],
              version: workerVersionsSample.results[0].id
            },
            workerRun2: {
              id: 'workerRun2',
              parents: ['workerRun1'],
              version: workerVersionsSample.results[1].id
            }
          }
        }

        mock.onDelete('/process/workers/workerRun1/').reply(204)
        await store.dispatch('process/deleteWorkerRun', { processId: 'processid', workerRunId: 'workerRun1' })
        assert.deepStrictEqual(store.history, [
          { action: 'process/deleteWorkerRun', payload: { processId: 'processid', workerRunId: 'workerRun1' } },
          {
            mutation: 'process/removeProcessWorkerRun',
            payload: {
              processId: 'processid',
              workerRunId: 'workerRun1'
            }
          }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'delete',
            url: '/process/workers/workerRun1/'
          }
        ])
        assert.deepStrictEqual(store.state.process.processWorkerRuns, { processid: { workerRun2: { id: 'workerRun2', parents: [], version: workerVersionsSample.results[1].id } } })
      })

      it('handles errors', async () => {
        mock.onDelete('/process/workers/workerRun1/').reply(500, { detail: ['Something went wrong'] })
        const error = await assertRejects(async () => store.dispatch('process/deleteWorkerRun', { processId: 'processid', workerRunId: 'workerRun1' }))
        assert.strictEqual(error.toString(), 'Error: Request failed with status code 500')
        assert.deepStrictEqual(store.history, [
          { action: 'process/deleteWorkerRun', payload: { processId: 'processid', workerRunId: 'workerRun1' } }
        ])
      })
    })

    describe('clearProcess', () => {
      it('removes all worker runs from a process', async () => {
        store.state.process.processWorkerRuns = { test_process_id: workerRunsSample, other_process_id: workerRunSample }
        mock.onDelete('/process/test_process_id/clear/').reply(204)
        await store.dispatch('process/clearProcess', { processId: 'test_process_id' })

        assert.deepStrictEqual(store.history, [
          { action: 'process/clearProcess', payload: { processId: 'test_process_id' } },
          {
            mutation: 'process/removeProcessWorkerRuns',
            payload: {
              processId: 'test_process_id'
            }
          }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'delete',
            url: '/process/test_process_id/clear/'
          }
        ])
        assert.deepStrictEqual(store.state.process.processWorkerRuns, { test_process_id: {}, other_process_id: workerRunSample })
      })
    })

    describe('startProcess', () => {
      it('starts a specific process', async () => {
        store.state.process.processes = processesSample

        const reply = { ...processesSample.process_id, state: 'running' }
        mock.onPost('/process/process_id/start/').reply(200, reply)

        const payload = { processId: 'process_id' }
        await store.dispatch('process/startProcess', payload)
        assert.deepStrictEqual(store.history, [
          { action: 'process/startProcess', payload },
          { mutation: 'process/setProcess', payload: reply }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'post',
            url: '/process/process_id/start/'
          }
        ])
        assert.deepStrictEqual(store.state.process.processes, { process_id: { ...reply, _complete: false } })
      })
    })

    describe('listProcesses', () => {
      it('list processes with different filters', async () => {
        store.state.process.processes = processesSample
        store.state.process.processPage = {}
        const reply = {
          count: 2,
          results: [
            { id: 'process_1' },
            { id: 'process_2' }
          ]
        }
        mock.onGet('/process/').reply(201, reply)

        await store.dispatch('process/listProcesses', { state: 'completed', page: '' })
        assert.deepStrictEqual(store.history, [
          { action: 'process/listProcesses', payload: { state: 'completed', page: '' } },
          { mutation: 'process/setProcesses', payload: reply.results },
          { mutation: 'process/setProcessPage', payload: reply }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'params'])), [
          {
            method: 'get',
            url: '/process/',
            params: { state: 'completed' }
          }
        ])
        assert.deepStrictEqual(store.state.process.processes, {
          ...processesSample,
          process_1: { id: 'process_1', _complete: false },
          process_2: { id: 'process_2', _complete: false }
        })
        assert.deepStrictEqual(store.state.process.processPage, {
          count: 2,
          results: ['process_1', 'process_2']
        })
      })

      it('handle errors listing processes', async () => {
        store.state.process.processes = processesSample
        store.state.process.processPage = { count: 1 }
        mock.onGet('/process/').reply(500, { detail: ['Something went wrong'] })

        await assertRejects(async () => store.dispatch('process/listProcesses', { state: 'completed', page: '' }))

        assert.deepStrictEqual(store.history, [
          { action: 'process/listProcesses', payload: { state: 'completed', page: '' } },
          { mutation: 'process/setProcessPage', payload: { results: [] } }
        ])
        assert.deepStrictEqual(store.state.process.processPage, {
          count: 1,
          results: []
        })
      })
    })

    describe('retryProcess', () => {
      it('retries a process', async () => {
        store.state.process.processes = {
          process_id: { id: 'process_id', state: 'completed' }
        }
        const reply = { id: 'process_id', state: 'unscheduled' }
        mock.onPost('/process/process_id/retry/').reply(200, reply)

        await store.dispatch('process/retryProcess', 'process_id')
        assert.deepStrictEqual(store.history, [
          { action: 'process/retryProcess', payload: 'process_id' },
          { mutation: 'process/setProcess', payload: reply }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'post',
            url: '/process/process_id/retry/',
            data: undefined
          }
        ])
        assert.deepStrictEqual(store.state.process.processes, {
          process_id: { id: 'process_id', state: 'unscheduled', _complete: false }
        })
      })
    })

    describe('createProcess', () => {
      it('creates a specific process from a corpus', async () => {
        store.state.process.processes = processesSample

        const reply = {
          id: 'new_process',
          state: 'unscheduled',
          mode: 'workers',
          corpus: 'corpus_2'
        }
        mock.onPost('/process/corpus/').reply(201, reply)

        await store.dispatch('process/createProcess', { corpus_id: 'corpus_2' })
        assert.deepStrictEqual(store.history, [
          { action: 'process/createProcess', payload: { corpus_id: 'corpus_2' } },
          { mutation: 'process/setProcess', payload: reply }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'post',
            url: '/process/corpus/',
            data: { corpus_id: 'corpus_2' }
          }
        ])
        assert.deepStrictEqual(store.state.process.processes, {
          ...processesSample,
          new_process: { ...reply, _complete: false }
        })
      })
    })

    describe('retrieveProcess', () => {
      it('retrieves a process details', async () => {
        store.state.process.processes.processid = {
          ...processSample,
          _complete: false
        }
        delete store.state.process.processes.processid.farm
        delete store.state.process.processes.processid.tasks

        const reply = { ...processSample, state: 'error' }
        const expectedProcess = {
          ...reply,
          _complete: true
        }
        delete expectedProcess.tasks

        mock.onGet('/process/processid/').reply(200, reply)

        await store.dispatch('process/retrieveProcess', 'processid')
        assert.deepStrictEqual(store.history, [
          { action: 'process/retrieveProcess', payload: 'processid' },
          { mutation: 'process/setProcess', payload: reply }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'get',
            url: '/process/processid/',
            data: undefined
          }
        ])
        assert.deepStrictEqual(store.state.process.processes, {
          processid: expectedProcess
        })
      })
    })

    describe('deleteProcess', () => {
      it('deletes a process', async () => {
        store.state.process.processes = processesSample

        mock.onDelete('/process/process_id/').reply(204)

        await store.dispatch('process/deleteProcess', 'process_id')
        assert.deepStrictEqual(store.history, [
          { action: 'process/deleteProcess', payload: 'process_id' },
          { mutation: 'process/removeProcess', payload: 'process_id' }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'delete',
            url: '/process/process_id/',
            data: undefined
          }
        ])
        assert.deepStrictEqual(store.state.process.processes, {})
      })
    })

    describe('listProcessElements', () => {
      it('lists elements involved in a process', async () => {
        store.state.process.processElementsPage = {
          process_0: { count: 1, results: [{ id: 'elt_0', type: 'line' }] },
          process_id: { count: null, results: [{ id: 'elt_id', type: 'page' }] }
        }

        const payload = { processId: 'process_id', cursor: 'ABC' }
        const response = { ...processElementsSample, count: 42 }
        // with_count attribute is automatically loaded when a page is missing count value
        mock.onGet('/process/process_id/elements/', { params: { cursor: 'ABC', with_count: true } }).reply(200, response)

        await store.dispatch('process/listProcessElements', payload)
        assert.deepStrictEqual(store.history, [
          { action: 'process/listProcessElements', payload },
          {
            mutation: 'process/setProcessElementsPage',
            payload: { processId: 'process_id', response }
          }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'params', 'data'])), [
          {
            method: 'get',
            url: '/process/process_id/elements/',
            params: {
              cursor: 'ABC',
              with_count: true
            },
            data: undefined
          }
        ])
        assert.deepStrictEqual(store.state.process.processElementsPage, {
          process_0: { count: 1, results: [{ id: 'elt_0', type: 'line' }] },
          process_id: response
        })
      })
    })

    describe('listProcessSets', () => {
      it('lists sets on a process', async () => {
        const set1 = {
          id: 'processsetid',
          dataset: { id: 'datasetid' },
          set_name: 'test'
        }
        const set2 = {
          id: 'processsetid2',
          dataset: { id: 'datasetid' },
          set_name: 'validation'
        }
        mock.onGet('/process/processid/sets/', { params: { page: 1 } }).reply(200, {
          count: 2,
          previous: null,
          next: '/process/processid/sets/?page=2',
          number: 1,
          results: [set1]
        })
        mock.onGet('/process/processid/sets/', { params: { page: 2 } }).reply(200, {
          count: 2,
          previous: '/process/processid/sets/?page=1',
          next: null,
          number: 2,
          results: [set2]
        })

        await store.dispatch('process/listProcessSets', { processId: 'processid' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'process/listProcessSets',
            payload: {
              processId: 'processid'
            }
          },
          {
            mutation: 'process/setProcessSets',
            payload: {
              processId: 'processid',
              results: [
                set1
              ]
            }
          },
          {
            action: 'process/listProcessSets',
            payload: {
              processId: 'processid',
              page: 2
            }
          },
          {
            mutation: 'process/setProcessSets',
            payload: {
              processId: 'processid',
              results: [
                set2
              ]
            }
          }
        ])

        assert.deepStrictEqual(store.state.process.processSets, {
          processid: [
            set1,
            set2
          ]
        })
      })

      it('detects pagination errors', async () => {
        const set1 = {
          id: 'processsetid',
          dataset: { id: 'datasetid' },
          set_name: 'test'
        }
        // All requests on any page number will return page 1
        mock.onGet('/process/processid/sets/').reply(200, {
          count: 2,
          previous: null,
          next: '/process/processid/sets/?page=2',
          number: 1,
          results: [set1]
        })

        const err = await assertRejects(async () => store.dispatch('process/listProcessSets', { processId: 'processid' }))

        assert.strictEqual(err.toString(), 'Error: Pagination failed while listing sets for process "processid"')

        assert.deepStrictEqual(store.history, [
          {
            action: 'process/listProcessSets',
            payload: {
              processId: 'processid'
            }
          },
          {
            mutation: 'process/setProcessSets',
            payload: {
              processId: 'processid',
              results: [
                set1
              ]
            }
          },
          {
            action: 'process/listProcessSets',
            payload: {
              processId: 'processid',
              page: 2
            }
          }
        ])

        assert.deepStrictEqual(store.state.process.processSets, {
          processid: [
            set1
          ]
        })
      })

      it('throws other errors', async () => {
        const set1 = {
          id: 'processsetid',
          dataset: { id: 'datasetid' },
          set_name: 'test'
        }
        mock.onGet('/process/processid/sets/', { params: { page: 1 } }).reply(200, {
          count: 2,
          previous: null,
          next: '/process/processid/sets/?page=2',
          number: 1,
          results: [set1]
        })
        mock.onGet('/process/processid/sets/', { params: { page: 2 } }).reply(418)

        const err = await assertRejects(async () => store.dispatch('process/listProcessSets', { processId: 'processid' }))

        assert.strictEqual(err.toString(), 'Error: Request failed with status code 418')

        assert.deepStrictEqual(store.history, [
          {
            action: 'process/listProcessSets',
            payload: {
              processId: 'processid'
            }
          },
          {
            mutation: 'process/setProcessSets',
            payload: {
              processId: 'processid',
              results: [
                set1
              ]
            }
          },
          {
            action: 'process/listProcessSets',
            payload: {
              processId: 'processid',
              page: 2
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              text: 'Request failed with status code 418',
              type: 'error'
            }
          }
        ])

        assert.deepStrictEqual(store.state.process.processSets, {
          processid: [
            set1
          ]
        })
      })

      it('does not list sets if they were already listed', async () => {
        // The array exists for this process, meaning a list already ran
        store.state.process.processSets.processid = []

        await store.dispatch('process/listProcessSets', { processId: 'processid' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'process/listProcessSets',
            payload: {
              processId: 'processid'
            }
          }
        ])

        assert.deepStrictEqual(store.state.process.processSets, {
          processid: []
        })
      })
    })

    describe('createProcessSet', () => {
      it('adds a dataset set to a process', async () => {
        mock.onPost('/process/processid/set/setid/').reply(201, {
          id: 'processsetid',
          dataset: { id: 'datasetid' },
          set_name: 'test'
        })

        await store.dispatch('process/createProcessSet', {
          processId: 'processid',
          setId: 'setid'
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'process/createProcessSet',
            payload: {
              processId: 'processid',
              setId: 'setid'
            }
          },
          {
            mutation: 'process/setProcessSets',
            payload: {
              processId: 'processid',
              results: [
                {
                  id: 'processsetid',
                  dataset: { id: 'datasetid' },
                  set_name: 'test'
                }
              ]
            }
          }
        ])

        assert.deepStrictEqual(store.state.process.processSets, {
          processid: [
            {
              id: 'processsetid',
              dataset: { id: 'datasetid' },
              set_name: 'test'
            }
          ]
        })
      })
    })

    describe('deleteProcessSet', () => {
      it('removes a set from a process', async () => {
        store.state.process.processSets.processid = [
          {
            id: 'processsetid',
            dataset: { id: 'datasetid' },
            set_name: 'test'
          }
        ]
        mock.onDelete('/process/processid/set/setid/').reply(201)

        await store.dispatch('process/deleteProcessSet', {
          processId: 'processid',
          setId: 'setid',
          processSetId: 'processsetid'
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'process/deleteProcessSet',
            payload: {
              processId: 'processid',
              setId: 'setid',
              processSetId: 'processsetid'
            }
          },
          {
            mutation: 'process/removeProcessSet',
            payload: {
              processId: 'processid',
              processSetId: 'processsetid'
            }
          }
        ])

        assert.deepStrictEqual(store.state.process.processSets, {
          processid: []
        })
      })

      it('notifies of errors', async () => {
        mock.onDelete('/process/processid/set/setid/').reply(418)

        const err = await assertRejects(async () => store.dispatch('process/deleteProcessSet', {
          processId: 'processid',
          setId: 'setid'
        }))

        assert.strictEqual(err.toString(), 'Error: Request failed with status code 418')

        assert.deepStrictEqual(store.history, [
          {
            action: 'process/deleteProcessSet',
            payload: {
              processId: 'processid',
              setId: 'setid'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 418'
            }
          }
        ])
      })
    })

    it('updateTask', async () => {
      const reply = {
        id: 'taskid',
        slug: 'task-slug',
        state: 'stopping',
        logs: 'hello'
      }
      mock.onPatch('/task/taskid/').reply(200, reply)
      await store.dispatch('process/updateTask', { id: 'taskid', state: 'stopping' })

      assert.deepStrictEqual(store.history, [
        {
          action: 'process/updateTask',
          payload: { id: 'taskid', state: 'stopping' }
        },
        {
          mutation: 'process/setTask',
          payload: reply
        }
      ])
      assert.deepStrictEqual(store.state.process.tasks, {
        taskid: {
          ...reply,
          timeoutId: null
        }
      })
    })

    it('getTaskArtifacts', async () => {
      mock.onGet('/task/taskid/artifacts/').reply(200, [artifactSample])

      await store.dispatch('process/getTaskArtifacts', 'taskid')

      assert.deepStrictEqual(store.history, [
        {
          action: 'process/getTaskArtifacts',
          payload: 'taskid'
        },
        {
          mutation: 'process/setTaskArtifacts',
          payload: {
            taskId: 'taskid',
            artifacts: [artifactSample]
          }
        }
      ])
      assert.deepStrictEqual(store.state.process.artifacts, { taskid: [artifactSample] })
    })

    describe('fromFiles', () => {
      it('starts a file import', async () => {
        mock.onPost('/process/fromfiles/').reply(201, processSample)

        const workflowPayload = {
          data: {
            files: [
              'fileid1',
              'fileid2'
            ]
          }
        }

        assert.deepStrictEqual(
          await store.dispatch('process/fromFiles', workflowPayload),
          processSample
        )

        assert.deepStrictEqual(store.history, [
          { action: 'process/fromFiles', payload: workflowPayload }
        ])
      })

      it('handles and throws errors', async () => {
        mock.onPost('/process/fromfiles/').reply(500, { detail: ['Something went wrong'] })

        const workflowPayload = {
          data: {
            files: [
              'fileid1',
              'fileid2'
            ]
          }
        }

        const err = await assertRejects(async () => store.dispatch('process/fromFiles', workflowPayload))
        assert.strictEqual(err.message, 'Request failed with status code 500')

        assert.deepStrictEqual(store.history, [
          {
            action: 'process/fromFiles',
            payload: {
              data: {
                files: [
                  'fileid1',
                  'fileid2'
                ]
              }
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              text: 'Something went wrong',
              type: 'error'
            }
          }
        ])
      })
    })

    it('stop', async () => {
      const reply = {
        ...processSample,
        state: 'stopping',
        tasks: [
          {
            ...taskSample,
            state: 'stopping'
          }
        ]
      }
      mock.onPatch('/process/processid/').reply(200, reply)

      const expectedProcess = {
        ...reply,
        _complete: true
      }
      delete expectedProcess.tasks

      const expectedTasks = {
        [taskSample.id]: {
          ...taskSample,
          timeoutId: null,
          process_id: 'processid',
          state: 'stopping'
        }
      }

      await store.dispatch('process/stop', processSample)

      assert.deepStrictEqual(store.history, [
        { action: 'process/stop', payload: processSample },
        { mutation: 'process/setProcess', payload: reply }
      ])

      assert.deepStrictEqual(mock.history.all.map(obj => pick(obj, ['method', 'url', 'data'])), [
        {
          method: 'patch',
          url: '/process/processid/',
          data: { state: 'stopping' }
        }
      ])

      assert.deepStrictEqual(store.state.process.processes.processid, expectedProcess)
      assert.deepStrictEqual(store.state.process.tasks, expectedTasks)
    })

    describe('startPolling', () => {
      it('starts process polling', async () => {
        setTimeout.returns(1337)
        await store.dispatch('process/startPolling', 'processid')
        assert.deepStrictEqual(store.history, [
          { action: 'process/startPolling', payload: 'processid' },
          { mutation: 'process/stopPolling' },
          { mutation: 'process/setPollingProcessId', payload: 'processid' },
          { mutation: 'process/setProcessTimeoutId', payload: 1337 }
        ])
        // Called once with a function and 0 as the delay
        assert.ok(setTimeout.calledOnceWithExactly(sinon.match.func, sinon.match(0)))
        assert.strictEqual(store.state.process.pollingProcessId, 'processid')
        assert.strictEqual(store.state.process.processTimeoutId, 1337)

        // Setup to test the polling function
        store.history = []
        mock.onGet('/process/processid/').reply(200, processSample)
        const poll = setTimeout.firstCall.firstArg
        setTimeout.reset()
        setTimeout.returns(42)

        await poll()

        assert.deepStrictEqual(store.history, [
          { action: 'process/retrieveProcess', payload: 'processid' },
          { mutation: 'process/setProcess', payload: processSample },
          { mutation: 'process/setProcessTimeoutId', payload: 42 }
        ])
        // New poll is scheduled
        assert.ok(setTimeout.calledOnceWithExactly(poll, 4000))
      })

      it('handles errors', async () => {
        mock.onGet('/process/processid/').reply(500)
        setTimeout.returns(1337)

        await store.dispatch('process/startPolling', 'processid')
        const poll = setTimeout.firstCall.firstArg
        await poll()

        assert.deepStrictEqual(store.history, [
          // Start the polling
          { action: 'process/startPolling', payload: 'processid' },
          { mutation: 'process/stopPolling' },
          { mutation: 'process/setPollingProcessId', payload: 'processid' },
          { mutation: 'process/setProcessTimeoutId', payload: 1337 },
          // Run the polling function
          { action: 'process/retrieveProcess', payload: 'processid' },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Error while fetching process: Request failed with status code 500'
            }
          },
          { mutation: 'process/setProcessTimeoutId', payload: 1337 }
        ])
        assert.ok(setTimeout.calledTwice)
      })

      it('aborts polling on HTTP 4xx', async () => {
        mock.onGet('/process/processid/').reply(404)
        setTimeout.returns(1337)

        await store.dispatch('process/startPolling', 'processid')
        const poll = setTimeout.firstCall.firstArg
        await poll()

        assert.deepStrictEqual(store.history, [
          // Start the polling
          { action: 'process/startPolling', payload: 'processid' },
          { mutation: 'process/stopPolling' },
          { mutation: 'process/setPollingProcessId', payload: 'processid' },
          { mutation: 'process/setProcessTimeoutId', payload: 1337 },
          // Run the polling function
          { action: 'process/retrieveProcess', payload: 'processid' },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Error while fetching process: Request failed with status code 404'
            }
          },
          { mutation: 'process/stopPolling' }
        ])
        assert.ok(setTimeout.calledOnce)
      })

      it('aborts polling before starting a HTTP request if it is stopped', async () => {
        setTimeout.returns(1337)

        await store.dispatch('process/startPolling', 'processid')
        await store.commit('process/stopPolling')
        const poll = setTimeout.firstCall.firstArg
        await poll()

        assert.deepStrictEqual(store.history, [
          // Start the polling
          { action: 'process/startPolling', payload: 'processid' },
          { mutation: 'process/stopPolling' },
          { mutation: 'process/setPollingProcessId', payload: 'processid' },
          { mutation: 'process/setProcessTimeoutId', payload: 1337 },
          // Run the polling function
          { mutation: 'process/stopPolling' }
        ])
        assert.ok(setTimeout.calledOnce)
        // No API requests at all
        assert.ok(!mock.history.all.length)
      })

      it('does not reschedule a poll after an HTTP request if it is stopped', async () => {
        // Stop the polling during the HTTP request, which can happen if a user browses somewhere else during the request
        mock.onGet('/process/processid/').reply(() => {
          store.commit('process/stopPolling')
          return [200, processSample]
        })
        setTimeout.returns(1337)

        await store.dispatch('process/startPolling', 'processid')
        const poll = setTimeout.firstCall.firstArg
        await poll()

        assert.deepStrictEqual(store.history, [
          // Start the polling
          { action: 'process/startPolling', payload: 'processid' },
          { mutation: 'process/stopPolling' },
          { mutation: 'process/setPollingProcessId', payload: 'processid' },
          { mutation: 'process/setProcessTimeoutId', payload: 1337 },
          // Run the polling function, causing a stop via the mock
          { action: 'process/retrieveProcess', payload: 'processid' },
          { mutation: 'process/stopPolling' },
          { mutation: 'process/setProcess', payload: processSample }
        ])
        assert.ok(setTimeout.calledOnce)
      })
    })

    describe('startTaskPolling', () => {
      it('starts task polling', async () => {
        store.state.process.tasks = {
          taskid: {
            id: 'taskid',
            state: 'running',
            timeoutId: null
          }
        }
        setTimeout.returns(1337)
        await store.dispatch('process/startTaskPolling', 'taskid')

        assert.deepStrictEqual(store.history, [
          { action: 'process/startTaskPolling', payload: 'taskid' },
          { mutation: 'process/stopTaskPolling', payload: 'taskid' },
          {
            mutation: 'process/setTaskTimeoutId',
            payload: {
              taskId: 'taskid',
              timeoutId: 1337
            }
          }
        ])
        // Called once with a function and 0 as the delay
        assert.ok(setTimeout.calledOnceWithExactly(sinon.match.func, sinon.match(0)))
        assert.strictEqual(store.state.process.tasks.taskid.timeoutId, 1337)

        // Setup to test the polling function
        store.history = []
        const taskReply = {
          id: 'taskid',
          state: 'completed'
        }
        mock.onGet('/task/taskid/from-agent/').reply(200, taskReply)
        const poll = setTimeout.firstCall.firstArg
        setTimeout.reset()
        setTimeout.returns(42)

        await poll()

        assert.deepStrictEqual(store.history, [
          { mutation: 'process/setTask', payload: taskReply },
          {
            mutation: 'process/setTaskTimeoutId',
            payload: {
              taskId: 'taskid',
              timeoutId: 42
            }
          }
        ])
        // New poll is scheduled
        assert.ok(setTimeout.calledOnceWithExactly(poll, 4000))
      })

      it('handles errors', async () => {
        store.state.process.tasks = {
          taskid: {
            id: 'taskid',
            slug: 'task_slug',
            state: 'running',
            timeoutId: null
          }
        }
        mock.onGet('/task/taskid/from-agent/').reply(500)
        setTimeout.returns(1337)

        await store.dispatch('process/startTaskPolling', 'taskid')
        const poll = setTimeout.firstCall.firstArg
        await poll()

        assert.deepStrictEqual(store.history, [
          { action: 'process/startTaskPolling', payload: 'taskid' },
          { mutation: 'process/stopTaskPolling', payload: 'taskid' },
          {
            mutation: 'process/setTaskTimeoutId',
            payload: {
              taskId: 'taskid',
              timeoutId: 1337
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Error while fetching task task_slug: Request failed with status code 500'
            }
          },
          {
            mutation: 'process/setTaskTimeoutId',
            payload: {
              taskId: 'taskid',
              timeoutId: 1337
            }
          }
        ])
        assert.ok(setTimeout.calledTwice)
      })

      it('aborts polling on HTTP 4xx', async () => {
        store.state.process.tasks = {
          taskid: {
            id: 'taskid',
            slug: 'task_slug',
            state: 'running',
            timeoutId: null
          }
        }
        mock.onGet('/task/taskid/from-agent/').reply(404)
        setTimeout.returns(1337)

        await store.dispatch('process/startTaskPolling', 'taskid')
        const poll = setTimeout.firstCall.firstArg
        await poll()

        assert.deepStrictEqual(store.history, [
          { action: 'process/startTaskPolling', payload: 'taskid' },
          { mutation: 'process/stopTaskPolling', payload: 'taskid' },
          {
            mutation: 'process/setTaskTimeoutId',
            payload: {
              taskId: 'taskid',
              timeoutId: 1337
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Error while fetching task task_slug: Request failed with status code 404'
            }
          },
          { mutation: 'process/stopTaskPolling', payload: 'taskid' }
        ])
        assert.ok(setTimeout.calledOnce)
      })

      it('aborts polling if the task is removed', async () => {
        store.state.process.tasks = {
          taskid: {
            id: 'taskid',
            slug: 'task_slug',
            state: 'running',
            timeoutId: null
          }
        }
        setTimeout.returns(1337)

        await store.dispatch('process/startTaskPolling', 'taskid')
        delete store.state.process.tasks.taskid

        const poll = setTimeout.firstCall.firstArg
        await poll()

        assert.deepStrictEqual(store.history, [
          { action: 'process/startTaskPolling', payload: 'taskid' },
          { mutation: 'process/stopTaskPolling', payload: 'taskid' },
          {
            mutation: 'process/setTaskTimeoutId',
            payload: {
              taskId: 'taskid',
              timeoutId: 1337
            }
          }
        ])
        assert.ok(setTimeout.calledOnce)
      })

      it('aborts polling before starting a HTTP request if it is stopped', async () => {
        store.state.process.tasks = {
          taskid: {
            id: 'taskid',
            slug: 'task_slug',
            state: 'running',
            timeoutId: null
          }
        }
        setTimeout.returns(1337)

        await store.dispatch('process/startTaskPolling', 'taskid')
        await store.commit('process/stopTaskPolling', 'taskid')
        const poll = setTimeout.firstCall.firstArg
        await poll()

        assert.deepStrictEqual(store.history, [
          { action: 'process/startTaskPolling', payload: 'taskid' },
          { mutation: 'process/stopTaskPolling', payload: 'taskid' },
          {
            mutation: 'process/setTaskTimeoutId',
            payload: {
              taskId: 'taskid',
              timeoutId: 1337
            }
          },
          { mutation: 'process/stopTaskPolling', payload: 'taskid' }
        ])
        assert.ok(setTimeout.calledOnce)
      })

      it('does not reschedule a poll after an HTTP request if it is stopped', async () => {
        store.state.process.tasks = {
          taskid: {
            id: 'taskid',
            slug: 'task_slug',
            state: 'running',
            timeoutId: null
          }
        }
        const taskReply = {
          id: 'taskid',
          state: 'completed'
        }
        mock.onGet('/task/taskid/from-agent/').reply(() => {
          store.commit('process/stopTaskPolling', 'taskid')
          return [200, taskReply]
        })
        setTimeout.returns(1337)

        await store.dispatch('process/startTaskPolling', 'taskid')
        const poll = setTimeout.firstCall.firstArg
        await poll()

        assert.deepStrictEqual(store.history, [
          { action: 'process/startTaskPolling', payload: 'taskid' },
          { mutation: 'process/stopTaskPolling', payload: 'taskid' },
          {
            mutation: 'process/setTaskTimeoutId',
            payload: {
              taskId: 'taskid',
              timeoutId: 1337
            }
          },
          { mutation: 'process/stopTaskPolling', payload: 'taskid' },
          { mutation: 'process/setTask', payload: taskReply }
        ])
        assert.ok(setTimeout.calledOnce)
      })
    })

    describe('selectFailures', () => {
      it('selects failed elements on a process', async () => {
        mock.onPost('/process/processId/select-failures/').reply(200)
        mock.onGet('/elements/selection/').reply(200, { results: [] })
        await store.dispatch('process/selectFailures', 'processId')
        assert.deepStrictEqual(store.history, [
          { action: 'process/selectFailures', payload: 'processId' },
          { action: 'selection/get', payload: {} },
          { mutation: 'selection/reset' },
          { mutation: 'notifications/notify', payload: { text: 'Elements with failures have been added to your selection', type: 'success' } }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          { method: 'post', url: '/process/processId/select-failures/' },
          { method: 'get', url: '/elements/selection/' }
        ])
      })

      it('displays an error when no failed element can be selected', async () => {
        mock.onPost('/process/processId/select-failures/').reply(400, { __all__: 'That process is not finished' })
        await store.dispatch('process/selectFailures', 'processId')
        assert.deepStrictEqual(store.history, [
          { action: 'process/selectFailures', payload: 'processId' },
          { mutation: 'notifications/notify', payload: { text: 'That process is not finished', type: 'error' } }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          { method: 'post', url: '/process/processId/select-failures/' }
        ])
      })
    })
  })
})
