import { assert } from 'chai'
import axios from 'axios'
import { FakeAxios, assertRejects, setUpTestPinia, actionsCompletedPlugin } from '../testhelpers'
import { useEntityStore, useNotificationStore } from '@/stores'
import { elementSample, entitiesSample, entitySample, entitiesInTranscriptionSample } from '../samples'

describe('entity', () => {
  describe('actions', () => {
    let mock, store, notificationStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setUpTestPinia([actionsCompletedPlugin])
      store = useEntityStore()
      notificationStore = useNotificationStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
      notificationStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('get', () => {
      it('retrieves an entity', async () => {
        mock.onGet('/entity/entityid/').reply(200, entitySample)
        await store.get('entityid')
        assert.deepStrictEqual(store.entity, entitySample)
      })

      it('handles errors', async () => {
        mock.onGet('/entity/entityid/').reply(500)
        await assertRejects(() => store.get('entityid'))
        assert.deepStrictEqual(store.entity, null)
      })
    })

    describe('list', () => {
      it('lists entities', async () => {
        mock.onGet('/corpus/corpusid/entities/').reply(200, entitiesSample)
        await store.list('corpusid')
        assert.deepStrictEqual(store.entities, entitiesSample)
      })

      it('handles errors', async () => {
        mock.onGet('/corpus/corpusid/entities/').reply(451, { detail: 'PC LOAD LETTER' })
        const error = await assertRejects(() => store.list('corpusid'))

        assert.deepStrictEqual(error.response.data, { detail: 'PC LOAD LETTER' })
        assert.deepStrictEqual(store.entities, null)
        assert.deepStrictEqual(notificationStore.notifications, [{
          id: 0,
          type: 'error',
          text: 'PC LOAD LETTER'
        }])
      })
    })

    describe('listElements', () => {
      it('lists elements related to an entity', async () => {
        mock.onGet('/entity/entityid/elements/').reply(200, elementSample)
        await store.listElements('entityid')
        assert.deepStrictEqual(store.elements, elementSample)
      })

      it('handles errors', async () => {
        mock.onGet('/entity/entityid/elements/').reply(500, 'oh no')
        await store.listElements('entityid')
        assert.deepStrictEqual(store.elements, null)
        assert.deepStrictEqual(notificationStore.notifications, [{
          id: 0,
          type: 'error',
          text: 'oh no'
        }])
      })
    })

    describe('listInTranscription', () => {
      it('lists entities in a transcription', async () => {
        mock.onGet('/transcription/transcriptionid/entities/').reply(200, entitiesInTranscriptionSample)
        await store.listInTranscription('transcriptionid')
        assert.deepStrictEqual(store.inTranscription, {
          transcriptionid:
          {
            ...entitiesInTranscriptionSample,
            loaded: 1,
            previous: null,
            number: 0
          }
        })
      })

      it('paginates through entities in a transcription', async () => {
        const firstPage = {
          ...entitiesInTranscriptionSample,
          next: '/transcription/transcriptionid/entities/?page=2'
        }
        mock.onGet('/transcription/transcriptionid/entities/', { params: { page: 1 } }).reply(200, firstPage)
        mock.onGet('/transcription/transcriptionid/entities/', { params: { page: 2 } }).reply(200, entitiesInTranscriptionSample)

        await store.listInTranscription('transcriptionid')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.inTranscription, {
          transcriptionid: {
            results: [
              ...entitiesInTranscriptionSample.results,
              ...entitiesInTranscriptionSample.results
            ],
            count: 1,
            loaded: 2,
            next: null,
            previous: null,
            number: 0
          }
        })
      })

      it('handles errors', async () => {
        const firstPage = {
          ...entitiesInTranscriptionSample,
          next: '/transcription/transcriptionid/entities/?page=2'
        }
        mock.onGet('/transcription/transcriptionid/entities/', { params: { page: 1 } }).reply(200, firstPage)
        mock.onGet('/transcription/transcriptionid/entities/', { params: { page: 2 } }).reply(500, 'oh no')

        await store.listInTranscription('transcriptionid')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.inTranscription, {
          transcriptionid:
          {
            ...entitiesInTranscriptionSample,
            loaded: 1,
            next: null,
            previous: null,
            number: 0,
            count: 1
          }
        })
        assert.deepStrictEqual(notificationStore.notifications, [{
          id: 0,
          text: 'oh no',
          type: 'error'
        }])
      })
    })
  })
})
