import axios from 'axios'
import { assert } from 'chai'
import { createPinia, setActivePinia } from 'pinia'
import { useNotificationStore, useIngestStore } from '@/stores'
import { bucketsSample } from '../samples'
import { assertRejects, FakeAxios } from '../testhelpers'

describe('ingest', () => {
  let mock, store, notificationStore

  before('Setting up mocks', () => {
    mock = new FakeAxios(axios)
    setActivePinia(createPinia())
    store = useIngestStore()
    notificationStore = useNotificationStore()
  })

  beforeEach(() => {
    // Remove any handlers, but leave mocking in place
    mock.reset()
    store.$reset()
  })

  after('Removing Axios mock', () => {
    // Remove mocking entirely
    mock.restore()
  })

  describe('listBuckets', () => {
    it('lists available S3 buckets', async () => {
      mock.onGet('/ingest/buckets/').reply(200, bucketsSample)

      await store.listBuckets()

      assert.deepStrictEqual(store.buckets, {
        'bucket number one': {
          name: 'bucket number one'
        },
        'secchio numero due': {
          name: 'secchio numero due'
        },
        'emmer nummer drie': {
          name: 'emmer nummer drie'
        }
      })
    })
  })

  describe('fromBucket', () => {
    it('starts a process from an S3 bucket', async () => {
      mock.onPost('/ingest/create/').reply(201, { id: 'ABCDEF' })
      const payload = { bucket_name: 'bucket number one', prefix: 'some/folder/' }

      assert.strictEqual(
        await store.fromBucket(payload),
        'ABCDEF'
      )
    })

    it('handles and throws errors', async () => {
      mock.onPost('/ingest/create/').reply(403, { detail: ['You do not have access to a thing.'] })
      const payload = { bucket_name: 'bucket number one', prefix: 'some/folder/' }

      const error = await assertRejects(async () => store.fromBucket(payload))

      assert.deepStrictEqual(error.response.data, { detail: ['You do not have access to a thing.'] })

      assert.deepStrictEqual(notificationStore.notifications, [
        {
          id: 0,
          type: 'error',
          text: 'You do not have access to a thing.'
        }
      ])
    })
  })
})
