import axios from 'axios'
import { assert } from 'chai'
import { pick } from 'lodash'
import { setActivePinia } from 'pinia'
import { workersSample } from '../samples'
import { useModelStore, useNotificationStore } from '@/stores'
import { FakeAxios, setUpTestPinia, actionsCompletedPlugin } from '../testhelpers'

describe('model', () => {
  describe('actions', () => {
    let mock, store, notificationStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      const [pinia] = setUpTestPinia([actionsCompletedPlugin])
      setActivePinia(pinia)
      store = useModelStore()
      notificationStore = useNotificationStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
      notificationStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('createModel', () => {
      it('creates a model', async () => {
        store.models = { model_0: { id: 'model_0' } }
        const attrs = { name: 'test', description: 'A' }
        mock.onPost('/models/').reply(201, { id: 'model_1', ...attrs })

        await store.createModel(attrs)

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'post',
            url: '/models/',
            data: attrs
          }
        ])
        assert.deepStrictEqual(store.models, {
          model_0: { id: 'model_0' },
          model_1: { id: 'model_1', ...attrs }
        })
      })
    })

    describe('updateModel', () => {
      it('updates a model', async () => {
        store.models = { modelid: { id: 'modelid', description: 'hello', archived: false } }
        const attrs = { name: 'test', description: 'A', archived: true }
        mock.onPatch('/model/modelid/').reply(201, { id: 'modelid', ...attrs })

        await store.updateModel('modelid', attrs)

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'patch',
            url: '/model/modelid/',
            data: attrs
          }
        ])
        assert.deepStrictEqual(store.models, {
          modelid: { id: 'modelid', name: 'test', description: 'A', archived: true }
        })
      })
    })

    describe('listModels', () => {
      it('lists stored models with filters', async () => {
        const reply = {
          count: 1,
          results: [
            { id: 'model1' }
          ]
        }
        mock.onGet('/models/').reply(200, reply)
        await store.listModels({ name: '1' })
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'params'])), [
          {
            method: 'get',
            url: '/models/',
            params: { name: '1' }
          }
        ])
        assert.deepStrictEqual(store.models, { model1: { id: 'model1' } })
      })
    })

    describe('retrieveModel', () => {
      it('retrieves a model', async () => {
        const model = { id: 'model1' }
        mock.onGet('/model/model1/').reply(200, model)

        await store.retrieveModel('model1')

        assert.deepStrictEqual(store.models, {
          model1: model
        })
      })
    })

    describe('listModelVersions', () => {
      it('lists stored model versions', async () => {
        const reply = {
          count: 1,
          results: [
            { id: 'model_version1' }
          ]
        }
        mock.onGet('/model/model1/versions/').reply(200, reply)
        await store.listModelVersions('model1', 1)
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'params'])), [
          {
            method: 'get',
            url: '/model/model1/versions/',
            params: { page: 1 }
          }])
        assert.deepStrictEqual(store.modelVersions, { model_version1: { id: 'model_version1' } })
      })
    })

    describe('getModelVersion', () => {
      it('retrieves a model version', async () => {
        const version = { id: 'model_version1' }
        mock.onGet('/modelversion/model_version1/').reply(200, version)

        await store.getModelVersion('model_version1')

        assert.deepStrictEqual(store.modelVersions, {
          model_version1: version
        })
      })
    })

    describe('deleteModelVersions', () => {
      it('deletes a model version', async () => {
        store.modelVersions = { model_version_0: { id: 'model_version_0' } }

        mock.onDelete('/modelversion/model_version_0/').reply(204)
        await store.deleteModelVersion('model_version_0')
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'delete',
            url: '/modelversion/model_version_0/'
          }])
        assert.deepStrictEqual(store.modelVersions, {})
      })

      it('handles errors', async () => {
        store.modelVersions = { model_version_0: { id: 'model_version_0' } }
        mock.onDelete('/modelversion/model_version_0/').reply(500)

        await store.deleteModelVersion('model_version_0')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'delete',
            url: '/modelversion/model_version_0/'
          }])
        assert.deepStrictEqual(store.modelVersions, { model_version_0: { id: 'model_version_0' } })
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'Request failed with status code 500'
          }
        ])
      })
    })

    describe('updateModelVersion', () => {
      it('updates the description and/or tag of a model version', async () => {
        store.modelVersions = {
          model_version_0: {
            id: 'model_version_0',
            description: "C'est un trou de verdure où chante une rivière"
          }
        }

        const updatedVersion = {
          id: 'model_version_0',
          description: "Accrochant follement aux herbes des haillons d'argent",
          tag: 'rimbaud'
        }

        mock.onPatch('/modelversion/model_version_0/').reply(200, updatedVersion)
        await store.updateModelVersion('model_version_0', { description: "Accrochant follement aux herbes des haillons d'argent", tag: 'rimbaud' })
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'patch',
            url: '/modelversion/model_version_0/'
          }])
        assert.deepStrictEqual(store.modelVersions, {
          model_version_0: {
            id: 'model_version_0',
            description: "Accrochant follement aux herbes des haillons d'argent",
            tag: 'rimbaud'
          }
        })
      })
    })

    describe('listAllModels', () => {
      it('lists all models (not paginated)', async () => {
        const pages = [
          {
            count: 4,
            number: 1,
            next: 'nextpage',
            results: [
              { id: 'model1' },
              { id: 'model2' },
              { id: 'model3' }
            ]
          },
          {
            count: 4,
            number: 2,
            next: null,
            results: [
              { id: 'model4' }
            ]
          }
        ]
        mock.onGet('/models/', { params: { page: 1 } }).reply(200, pages[0])
        mock.onGet('/models/', { params: { page: 2 } }).reply(200, pages[1])

        await store.listAllModels()

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'params'])), [
          {
            method: 'get',
            url: '/models/',
            params: { page: 1 }
          },
          {
            method: 'get',
            url: '/models/',
            params: { page: 2 }
          }
        ])
        assert.deepStrictEqual(store.allModels, {
          model1: { id: 'model1' },
          model2: { id: 'model2' },
          model3: { id: 'model3' },
          model4: { id: 'model4' }
        })
      })
    })

    describe('listCompatibleWorkers', () => {
      it('list compatible workers for a model through all pages', async () => {
        store.compatibleWorkers = {
          model_id: [{ id: 'worker_1', name: 'Worker 1' }],
          other_model: [{ id: 'worker_1', name: 'Worker 1' }]
        }
        mock.onGet('/workers/', { params: { compatible_model: 'model_id', page: 1 } }).reply(200, {
          count: 2,
          number: 1,
          previous: null,
          next: 'next',
          results: [{ id: 'worker_2', name: 'Worker 2' }]
        })
        mock.onGet('/workers/', { params: { compatible_model: 'model_id', page: 2 } }).reply(200, {
          count: 2,
          number: 2,
          previous: 'previous',
          next: null,
          results: [{ id: 'worker_3', name: 'Worker 3' }]
        })
        await store.listCompatibleWorkers('model_id')
        await store.actionsCompleted()
        assert.deepStrictEqual(store.compatibleWorkers, {
          other_model: [{ id: 'worker_1', name: 'Worker 1' }],
          model_id: [{ id: 'worker_2', name: 'Worker 2' }, { id: 'worker_3', name: 'Worker 3' }]
        })
      })
    })

    describe('addCompatibleWorker', () => {
      it('sets a worker as compatible with a model', async () => {
        store.compatibleWorkers = {
          model_id: [{ id: 'worker_1', name: 'Worker 1' }]
        }
        const worker = workersSample.results[0]
        mock.onPost(`/model/model_id/compatible-worker/${worker.id}/`).reply(201)
        await store.addCompatibleWorker('model_id', worker)
        assert.deepStrictEqual(store.compatibleWorkers, {
          model_id: [{ id: 'worker_1', name: 'Worker 1' }, worker]
        })
      })
    })

    describe('removeCompatibleWorker', () => {
      it('unsets a worker as compatible with a model', async () => {
        const worker = workersSample.results[0]
        store.compatibleWorkers = {
          model_id: [
            { id: 'worker_1', name: 'Worker 1' },
            worker
          ]
        }
        mock.onDelete(`/model/model_id/compatible-worker/${worker.id}/`).reply(204)
        await store.removeCompatibleWorker('model_id', worker)
        assert.deepStrictEqual(store.compatibleWorkers, {
          model_id: [{ id: 'worker_1', name: 'Worker 1' }]
        })
      })
    })
  })
})
